from requests import get,head,post,delete as htdelete
from urllib.parse import urljoin
from contextlib import closing
from json import dumps,loads
from sys import stderr
from time import sleep
from kblab.result import create_result
from collections.abc import Iterator,Generator
import kblab
import logging

MAX_ID=2**38
MAX_RETRIES=2
MAX_RETRY_WAIT=60
VERIFY_CA=True

logging.getLogger('requests').setLevel(logging.CRITICAL)
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

logging.getLogger('requests').setLevel(logging.CRITICAL)
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()

class HttpArchive(kblab.Archive):
    def __init__(self, root=None, base=None, auth=None):
        url=root
        self.url = url + ('/' if url[-1] != '/' else '')
        self.base = base or url

        if auth:
            if isinstance(auth, dict) and auth['type'] == 'basic':
                self.auth = (auth['user'], auth['pass'])
            elif isinstance(auth, tuple) and len(auth) == 2:
                self.auth = auth
            else:
                raise Exception('unsupported auth configuration')
        else:
            self.auth = None

        n_retries=0
        retry_wait=1
        while True:
            try:
                r = get(urljoin(self.url, 'base'), auth=self.auth, verify=kblab.VERIFY_CA)
            except Exception as e:
                if n_retries == MAX_RETRIES:
                    raise Exception('Max number of retries (%d) reached. %s' % (MAX_RETRIES, str(e)))

                print('Exception while connecting. Retry %d of %s in %d seconds. %s' % (n_retries+1, str(MAX_RETRIES), retry_wait, str(e)), file=stderr, flush=True)
            else:
                if r.status_code in [ 401, 403 ]:
                    raise Exception('Unauthorized. HTTP status = %d' % r.status_code)

                if r.status_code == 200:
                    self.server_base = r.text
                    break;

                if n_retries == MAX_RETRIES:
                    raise Exception('Max number of retries (%d) reached.')
                
                print('Unexpected HTTP status code = %d Retry %d of %s in %d seconds' % (r.status_code, n_retries+1, str(MAX_RETRIES), retry_wait), file=stderr, flush=True)

            n_retries += 1
            sleep(retry_wait)

            if 2*retry_wait < MAX_RETRY_WAIT:
                retry_wait *= 2


    def get(self, key, mode='r'):
        try:
            return kblab.Package(
                    self.url + key + '/',
                    mode=mode,
                    base=urljoin(self.base, key + '/'),
                    auth=self.auth,
                    server_base=urljoin(self.server_base, key + '/'))
        except kblab.HttpNotFoundException:
            return None
        except:
            raise


    def location(self, key, path=None):
        return self.url + key + '/' + (path or '')


    def exists(self, key, path=None):
        url = self.location(key, path)

        with closing(head(url, auth=self.auth)) as r:
            if r.status_code == 200:
                return True
            elif r.status_code == 404:
                return False

            raise Exception(f'Server returned {r.status_code}')


    def open(self, key, path, mode=''):
        url = self.location(key, path)
        r = get(url, auth=self.auth, stream=True)

        if r.status_code == 200:
            r.raw.decode_stream = True

            return r.raw
        elif r.status_code == 404:
            raise FileNotFoundError(url)

        raise Exception(f'Server returned {r.status_code}')


    def read(self, key, path, mode=''):
        with self.open(key, path, mode=mode) as f:
            return f.read()


    def search(self, query, start=0, max=None, sort=None, level=None, include=False):
        if isinstance(query, dict):
            query = dumps(query)

        key_iter = self._search_iter(query, start, max, sort, level, include)
        start,max,count = [ int(x) for x in next(key_iter).split() ]

        return kblab.Result(
                    start,
                    min(count-start, max) if max else count-start,
                    count,
                    key_iter,
                    iter([]))


    def _search_iter(self, query, start=0, max=None, sort=None, level=None, include=False):
        params = { 'q': query, 'from': start }
        if max: params.update({ 'max': max })
        if level: params.update({ '@type': level })
        if sort: params.update({ 'sort': sort })
        if include: params.update({ 'include': 'True' })

        with closing(get(urljoin(self.url, '_search'),
            params=params,
            verify=kblab.VERIFY_CA,
            auth=self.auth,
            stream=True)) as r:

            if r.status_code == 200:
                first=True
                for line in r.raw:
                    line = line[:-1].decode('utf-8').split(',', 1)
                    
                    yield (line[0], loads(line[1])) if include and not first else line[0]

                    first=False
            else:
                raise Exception('Expected status 200, got %d' % r.status_code)


    def serialize(self, key_or_iter, resolve=True, iter_content=False, buffer_size=100*1024):
        if not isinstance(key_or_iter, str):
            raise Exception('Only single string as key_or_iter supported.')

        url = self.location(key_or_iter) + '_serialize'

        r = get(url, auth=self.auth, stream=True)
    
        if r.status_code == 200:
            r.raw.decode_stream = True

            if iter_content:
                return kblab.utils.stream_to_iter(r.raw, chunk_size=buffer_size)
            else:
                return r.raw
        elif r.status_code == 404:
            raise FileNotFoundError(url)

        raise Exception(f'Server returned {r.status_code}')


    def count(self, query, cats={}):
        with closing(get(urljoin(self.url, 'count'),
                         verify=kblab.VERIFY_CA,
                         params={ 'q': dumps(query), 'c': dumps(cats) },
                         auth=self.auth)) as r:
                
            if r.status_code == 200:
                return loads(r.text)
            else:
                raise Exception('Expected status 200, got %d' % r.status_code)


    def get_location(self, key, path):
        return self.url + key + '/' + path


    def keys(self):
        with closing(get(self.url + 'packages', auth=self.auth, stream=True, verify=kblab.VERIFY_CA)) as r:
            if r.status_code == 200:
                for key in r.raw:
                    yield key[:-1].decode('utf-8')
            elif r.status_code != 404:
                raise Exception('server returned status %d' % r.status_code)


    def __iter__(self):
        yield from self.keys()


    def items(self):
        for key in self:
            yield (key, self.get(key))


    def _get_key(self, url):
        if url.startswith(self.url):
            return url[len(self.url):].split('/')[0]

        raise Exception('url (%s) does not start with %s' % (url, self.url))


    def __contains__(self, key):
        with closing(get(self.url + key + '/', auth=self.auth, verify=kblab.VERIFY_CA)) as r:
            return r.status_code == 200


    def __getitem__(self, key):
        return self.get(key)





#!/usr/bin/env python

from uuid import uuid4
from hashlib import md5
from re import split
from contextlib import contextmanager
from os import makedirs
from os.path import dirname,exists,abspath
from hashlib import sha1
from io import BytesIO
from urllib.request import urlopen
from random import random
from datetime import datetime
from re import match
from copy import deepcopy
from collections import Counter
from sys import stderr
from json import load
from copy import copy
from subprocess import Popen

from io import StringIO

TEMP_PREFIX='/tmp/starch-temp-'

def __init__():
    pass

def get_temp_dirname():
    return TEMP_PREFIX + '%s/' % uuid4()


def convert(n, radix=32, pad_to=0, alphabet='0123456789bcdfghjklmnpqrstvxzBCDFGHJKLMNPQRSTVXZ'):
    ret = ''

    if radix > len(alphabet) or radix == 0:
        raise Exception('radix == 0 or radix > len(alphabet)')

    while n:
        n, rest = divmod(n, radix)
        ret = alphabet[rest] + ret

    while len(ret) < pad_to:
        ret = alphabet[0] + ret

    return ret or alphabet[0]


def valid_path(path):
    if path[0] in [ '/' ] or '../' in path:
        raise Exception('invalid path (%s)' % path)

    return path


def valid_file(path):
    if path[0] == '/' or '..' in path:
        raise Exception('invalid path (%s)' % path)

    return path


def valid_key(key):
    if '/' in key or '.' in key:
        raise Exception('invalid key (%s)' % key)

    return key


def timestamp():
    return datetime.utcnow().isoformat() + 'Z'


def dict_search(a, b):
    if isinstance(a, (str, int, float)):
        if isinstance(b, (str, int, float)):
            return a == b
        elif isinstance(b, (list, iter)):
            return any([ dict_search(a, x) for x in b ])
    elif isinstance(a, dict) and isinstance(b, dict):
        if set(a).issubset(b):
            return all([ dict_search(a[key], b[key]) for key in a ])

    return False


def wildcard_match(a, b):
    return a == b or a == '*'


def dict_values(d, path):
    p,di = [], deepcopy(path)

    while isinstance(di, dict):
        p += [ next(iter(di.keys())) ]
        di = next(iter(di.values()))

    p += [ di ]

    return _dict_values(d, p)


def _dict_values(d, path):
    if path == []:
        if isinstance(d, list):
            return Counter([ x for x in d if not isinstance(x, (list, dict, tuple)) ])
        elif not isinstance(d, (list, dict, tuple)):
            return Counter([d])
    else:
        if isinstance(d, (list, tuple)):
            s=Counter()
            for x in d:
                s.update(_dict_values(x, path))
            return s
        elif isinstance(d, dict) and path[0] in d:
            return _dict_values(d[path[0]], path[1:])
    
    return Counter()


def chunked(f, chunk_size=100*1024, max=None):
    pos,b = 0,None
    while b != b'' and b != '':
        b = f.read(chunk_size if not max else min(chunk_size, max - pos))
        #print('chunk %d' % len(b), file=stderr, flush=True)
        pos += len(b)

        if b != b'' and b != '':
            yield b


def stream_to_iter(raw, chunk_size=10*1024, max=None):
    with raw as f:
        yield from chunked(f, chunk_size=chunk_size, max=max)


def decode_range(srange):
    s = match('bytes=(\\d+)-(\\d*)', srange).groups()

    return ( int(s[0]), int(s[1]) if s[1] != '' else None )


@contextmanager
def nullctxmgr():
    yield


def nullcallback(msg, **kwargs):
    #print(msg, kwargs)

    if msg == 'lock':
        return nullctxmgr()


def rebase(desc, base, index_base, in_place=False):
    if base or index_base:
        ret = deepcopy(desc) if in_place else desc
        ret['@id'] = rebase_uri(['@id'], base, index_base)

        for f in ret['files']:
            f['@id'] = rebase_uri(f['@id'])

        return ret
    else:
        return desc


def rebase_uri(u, base, index_base):
    return (base or '') + (u[len(index_base):] if index_base else u)

def max_iter(i, max):
    n=0
    for chunk in i:
        if len(chunk) + n < max:
            yield chunk
            n += len(chunk)
        else:
            yield chunk[:max-n]

            return


def guess_content(path):
    if path.endswith('.jpg'):
        return 'image/jpeg'
    elif path.endswith('.jp2'):
        return 'image/jp2'
    elif path.endswith('.json'):
        return 'application/json'
    elif path.endswith('.pdf'):
        return 'application/pdf'
    elif path.endswith('.xml'):
        return 'text/xml'
    elif path.endswith('.metadata'):
        return 'text/xml'
    
    return 'application/octet-stream'


def flatten_structure(structure, levels=[ 'Part', 'Page' ]):
    for s in structure:
        if s['@type'] in levels:
            yield s

        if 'has_part' in s:
            yield from flatten_structure(s['has_part'])


def aggregate(structure, content, environment, meta):
    ret = copy(environment)
    ret_content = []

    if meta:
        ret['meta'] = meta

    stack = [ structure ]
    while len(stack) > 0:
        element = stack.pop(0)

        for x in element if isinstance(element, list) else [ element ]:
            i = x.get('@id', '')

            if i in content and content[i].get('content', '') != '':
                ret_content += [ content[i]['content'] if environment.get('@type', '') == 'Text' else content[i] ]

            if 'has_part' in x:
                stack.insert(0, x['has_part'])

    ret['content'] = ret_content

    return ret


def _flerge(structure, content, meta, level='Text', ignore = []):
    ret = []
    ignore += [ 'has_part', 'content' ]

    #stack = [ (deepcopy(meta), element) for element in structure ]
    stack = [ ({ 'path': []}, element) for element in structure ]
    while len(stack) != 0:
        environment,element = stack.pop(0)
        nenviron = { key:copy(environment[key]) for key in sorted(environment) }
        nenviron.update({ key:value for key,value in element.items() if key not in ignore })        

        if element.get('@type', None) == level:
            ret += [ aggregate(element, content, nenviron, meta) ]
        elif 'has_part' in element:
            nenviron['path'] += [ { '@id': element['@id'], '@type': element['@type'] } ]
            for x in element['has_part'][::-1]:
                stack.insert(0, (nenviron, x))

    return ret


def flerge(package, level='Text', ignore=[]):
    structure = load(package.get_raw('structure.json'))
    content = { x['@id']:x for x in load(package.get_raw('content.json')) }
    meta = load(package.get_raw('meta.json')) if 'meta.json' in package else {}
    desc = package.description()

    if level == 'Package':
        s = [ desc ]
        s[0]['meta'] = meta
        s[0]['has_part'] = structure
        
        structure = s
        #structure = [ { '@id': desc['@id'], '@type': desc['@type'], 'label': desc['label'], 'tags': desc['tags'], 'meta': meta, 'files': desc['files'], 'size': desc['size'], 'has_part': structure } ]
    else:
        structure = [ { '@id': desc['@id'], '@type': desc['@type'], 'label': desc['label'], 'tags': desc['tags'], 'meta': meta, 'has_part': structure } ]

    return _flerge(structure, content, meta, level, ignore)

